# Fish-like autosuggestions for zsh based on the code for 'predict-on'
#
# ```zsh
# zle-line-init() {
#		enable-autosuggestions
# }
# zle -N zle-line-init
# ```
zmodload zsh/net/socket

AUTOSUGGEST_SERVER_SCRIPT="${0:a:h}/completion-server.zsh"

# function {
# 	[[ -n $ZLE_DISABLE_AUTOSUGGEST ]] && return
# 	setopt local_options no_hup

# 	local server_dir="/tmp/zsh-autosuggest-$USER"
# 	local pid_file="$server_dir/pid"
# 	local socket_path="$server_dir/socket"

# 	if ! [[ -S $socket_path && -r $pid_file ]] || ! kill -0 $(<$pid_file) &>/dev/null; then
# 		# start server
# 		zsh $AUTOSUGGEST_SERVER_SCRIPT $server_dir $pid_file $socket_path &!
# 	fi

# 	integer remaining_tries=10
# 	# wait until the process is listening
# 	while ! [[ -d $server_dir && -r $pid_file ]] || ! kill -0 $(<$pid_file) &> /dev/null; do
# 		(( --remaining_tries )) || break
# 		sleep 0.3
# 	done

# 	ZLE_AUTOSUGGEST_SOCKET=$socket_path
# }

ZLE_AUTOSUGGEST_PAUSE_WIDGETS=(
vi-cmd-mode vi-backward-char backward-char backward-word beginning-of-line
history-search-forward history-search-backward up-line-or-history
down-line-or-history
)

ZLE_AUTOSUGGEST_COMPLETION_WIDGETS=(
complete-word expand-or-complete expand-or-complete-prefix list-choices
menu-complete reverse-menu-complete menu-expand-or-complete menu-select
accept-and-menu-complete
)

pause-autosuggestions() {
	[[ -n $ZLE_AUTOSUGGESTING_PAUSED ]] && return
	local widget
	# When autosuggestions are disabled, kill the unmaterialized part
	RBUFFER=''
	unset ZLE_AUTOSUGGESTING
	ZLE_AUTOSUGGESTING_PAUSED=1
	zle -A self-insert paused-autosuggest-self-insert
	zle -A .magic-space magic-space
	zle -A .backward-delete-char backward-delete-char
	zle -A .accept-line accept-line
	for widget in $ZLE_AUTOSUGGEST_PAUSE_WIDGETS; do
		eval "zle -A autosuggest-${widget}-orig ${widget}"
	done
	for widget in $ZLE_AUTOSUGGEST_COMPLETION_WIDGETS; do
		eval "zle -A autosuggest-${widget}-orig $widget"
	done
	highlight-suggested-text
}

enable-autosuggestions() {
	[[ -n $ZLE_AUTOSUGGESTING ]] && return
	local widget
	unset ZLE_AUTOSUGGESTING_PAUSED
	ZLE_AUTOSUGGESTING=1
	# Replace prediction widgets by versions that will also highlight RBUFFER
	zle -N self-insert autosuggest-self-insert
	zle -N magic-space autosuggest-self-insert
	zle -N backward-delete-char autosuggest-delete
	zle -N accept-line autosuggest-accept-line
	# Hook into some default widgets that should pause autosuggestion
	# automatically 
	for widget in $ZLE_AUTOSUGGEST_PAUSE_WIDGETS; do
		eval "zle -A $widget autosuggest-${widget}-orig; \
			zle -A autosuggest-pause $widget"
	done
	# Hook into completion widgets to handle suggestions after completions
	for widget in $ZLE_AUTOSUGGEST_COMPLETION_WIDGETS; do
		eval "zle -A $widget autosuggest-${widget}-orig; \
			zle -A autosuggest-tab $widget"
	done
	if [[ $BUFFER != '' ]]; then
		show-suggestion
	fi
}

disable-autosuggestions() {
	if [[ -z $ZLE_AUTOSUGGESTING_PAUSED ]]; then
		pause-autosuggestions
	fi
	unset ZLE_AUTOSUGGESTING_PAUSED
	zle -A .self-insert self-insert
}

# Toggles autosuggestions on/off
toggle-autosuggestions() {
	if [[ -n $ZLE_AUTOSUGGESTING || -n $ZLE_AUTOSUGGESTING_PAUSED ]]; then
		disable-autosuggestions
	else
		enable-autosuggestions
	fi
}

highlight-suggested-text() {
	if [[ -n $ZLE_AUTOSUGGESTING ]]; then
		local color='fg=8'
		[[ -n $AUTOSUGGESTION_HIGHLIGHT_COLOR ]] &&\
			color=$AUTOSUGGESTION_HIGHLIGHT_COLOR
		region_highlight=("$(( $CURSOR + 1 )) $(( $CURSOR + $#RBUFFER )) $color")
	else
		region_highlight=()
	fi
}

autosuggest-self-insert() {
	setopt localoptions noshwordsplit noksharrays
	if [[ ${RBUFFER[1]} == ${KEYS[-1]} ]]; then
		# Same as what's typed, just move on
		((++CURSOR))
		highlight-suggested-text
	else
		LBUFFER="$LBUFFER$KEYS"
		show-suggestion
	fi
}

# Taken from predict-on
autosuggest-delete() {
	if (( $#LBUFFER > 1 )); then
		setopt localoptions noshwordsplit noksharrays
		# When editing a multiline buffer, it's unlikely prediction is wanted;
		# or if the last widget was e.g. a motion, then probably the intent is
		# to actually edit the line, not change the search prefix.
		if [[ $LASTWIDGET != (self-insert|magic-space|backward-delete-char) ]]; then
			LBUFFER="$LBUFFER[1,-2]"
		else
			((--CURSOR))
			zle .history-beginning-search-forward || RBUFFER=""
			return 0
		fi
	else
		zle .kill-whole-line
	fi
	highlight-suggested-text
}

# When autosuggesting, ignore RBUFFER which corresponds to the 'unmaterialized'
# section when the user accepts the line
autosuggest-accept-line() {
	RBUFFER=''
	region_highlight=()
	zle .accept-line
}

paused-autosuggest-self-insert() {
	if [[ $RBUFFER == '' ]]; then
		# Resume autosuggestions when inserting at the end of the line
		enable-autosuggestions
		autosuggest-self-insert
	else
		zle .self-insert
	fi
}

autosuggest-first-completion() {
	zsocket $ZLE_AUTOSUGGEST_SOCKET &>/dev/null || return 1
	local connection=$REPLY
	local completion
	print -u $connection $LBUFFER
	while read -u $connection completion; do
		RBUFFER=" $completion"
		break
	done
	exec {connection}>&-
}

show-suggestion() {
	[[ -n $ZLE_DISABLE_AUTOSUGGEST || $LBUFFER == '' ]] && return
	# TODO need a way to reset HISTNO so .history-beginning-search-backward
	# will always retrieve the last matching history entry
	# unset HISTNO
	zle .history-beginning-search-backward || autosuggest-first-completion\
		|| RBUFFER=''
	highlight-suggested-text
}

autosuggest-pause() {
	pause-autosuggestions
	zle autosuggest-${WIDGET}-orig "$@"
}

autosuggest-tab() {
	RBUFFER=''
	zle autosuggest-${WIDGET}-orig "$@"
}

accept-suggested-small-word() {
	zle .vi-forward-word
	highlight-suggested-text
}

accept-suggested-word() {
	zle .forward-word
	highlight-suggested-text
}

zle -N toggle-autosuggestions
zle -N accept-suggested-small-word
zle -N accept-suggested-word
zle -N autosuggest-pause
zle -N autosuggest-tab
